#include <iostream>
#include <complex>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "fltfile.h"
#include "wavfile.h"

#ifdef GPL
	#include <fftw.h>
	#include <getopt.h>
#endif
#ifdef HARKLICENSE
	#include "fft.hpp"
	#include "getopt.h"
#endif

#ifdef DEBUG
	#define IFDEBUG(statement) statement
#else
	#define IFDEBUG(statement)
#endif

#ifdef cplusplus
extern "C" {
#endif
void fft(double *in_buf, double *out_buf_re, double *out_buf_im, int data_size);
void ifft(double *in_buf_re, double *in_buf_im, double *out_buf_re, double *out_buf_im, int data_size);
#ifdef cplusplus
}
#endif

void usage()
{
	printf("./calcSysImp [option] inputfile outimpfilehead \n");
	printf("calculate impluse from TSP wave\n");
	printf("(fix time: 2013/10/30, compile time: %s %s)\n",__DATE__ ,__TIME__);
	printf("\n");
	printf("DESCRIPTION\n");
	printf("\tinputfile   : 16bit or 24bit , 8ch or 10ch or 16ch , 16kHz wave format\n");
	printf("\t              48KHz is not available yet.\n");
	printf("\timpfilehead : its extention is k.flt where k is ch number.\n");
	printf("\t              each imp file is 32bit float mono 1-ch raw data\n");
	printf("\t  eg.) when outimpfilehead = 'a', file names are 'a_ch1.flt', 'a_ch2.flt', ...\n");
	printf("\n");
	printf("[option]\n");
	printf("\t--tsporigflt file_name   : playing tsp file(float-data) when recording tsp wave.\n");
	printf("\t--tsporigwav file_name   : playing tsp file(PCM-wave) when recording tsp wave.\n");
	printf("\t--tsporigchannel N       : select use cahnnel of tsp origin file(PCM-wave).\n");
	printf("\t--sigmax N               : set optional signal max for tsp original file\n");
	printf("\t                         (not working on using --tsporigwav.)\n");
	printf("\t--samplebyframe N        : sample by one frame(default = 16384) \n");
	printf("\t--syncaddfreq N          : frequency of synchronous addition(default = 16) \n");
	printf("\t--syncaddoffset N        : start position of synchronous addition(default = 16384) \n");
	printf("\t--searchoffset N,N       : using searched start position.(search start pos, search end pos is required.) \n");
	printf("\t--outchannellist N,N,... : select calulating channel.(first channel is ch0.)\n");
	printf("\t--enableHPF Lfreq,Hfreq  : enable HighPassFilter.\n");
		
	exit(-1);
	/*./calcSysImp ../TSP000.wav ../impulse/D000 */

}

void fft(double *in_buf, double *out_buf_re, double *out_buf_im, int data_size) {
#ifdef GPL
	fftw_complex *in, *out;
	fftw_plan p;
	int idx;

	IFDEBUG(printf("data_size %ld, data_count %ld\n", (long) (sizeof(fftw_complex) * data_size), (long) (data_size)));
	in = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
	out = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
	p = fftw_create_plan(data_size, FFTW_FORWARD, FFTW_ESTIMATE); // plan の作成

	// 入力データの設定
	for (idx = 0; idx < data_size; idx++) {
		in[idx].re = in_buf[idx];
		in[idx].im = 0;
	}

	IFDEBUG(printf("calc_fft...\n"));
	fftw_one(p, in, out); // fft の実行
	IFDEBUG(printf("calced_fft...\n"));

	// 出力データの設定
	for (idx = 0; idx < data_size; idx++) {
		out_buf_re[idx] = out[idx].re;
		out_buf_im[idx] = out[idx].im;
	}

	fftw_destroy_plan(p); // plan の破棄
	fftw_free(in);
	fftw_free(out);
#endif
#ifdef HARKLICENSE
	HRI::fft<double> fft_obj(data_size);
	for (int fft = 0; fft < data_size; fft++) {
		fft_obj.a[fft] = std::complex<double>(*(in_buf + fft), 0);
	}
	fft_obj.analyze();
	for (int fft = 0; fft < data_size; fft++) {
		out_buf_re[fft] = fft_obj.A[fft].real();
		out_buf_im[fft] = fft_obj.A[fft].imag();
	}

#endif
}

void ifft(double *in_buf_re, double *in_buf_im, double *out_buf_re, double *out_buf_im, int data_size) {
#ifdef GPL
	fftw_complex *in, *out;
	fftw_plan p;
	int idx;

	IFDEBUG(printf("data_size %ld, data_count %ld\n", (long) (sizeof(fftw_complex) * data_size), (long) data_size));
	in = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
	out = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
	p = fftw_create_plan(data_size, FFTW_BACKWARD, FFTW_ESTIMATE); // plan の作成

	// 入力データの設定
	for (idx = 0; idx < data_size; idx++) {
		in[idx].re = in_buf_re[idx];
		in[idx].im = in_buf_im[idx];
	}

	IFDEBUG(printf("calc_ifft...\n"));
	fftw_one(p, in, out); // fft の実行
	IFDEBUG(printf("calced_ifft...\n"));

	// 出力データの設定
	for (idx = 0; idx < data_size; idx++) {
		out_buf_re[idx] = out[idx].re;
		out_buf_im[idx] = out[idx].im;
	}

	fftw_destroy_plan(p); // plan の破棄
	fftw_free(in);
	fftw_free(out);
#endif
#ifdef HARKLICENSE
	HRI::fft<double> fft_obj(data_size);
	for (int fft = 0; fft < data_size; fft++) {
		fft_obj.A[fft] = std::complex<double>(*(in_buf_re+fft), *(in_buf_im+fft));
	}
	fft_obj.synthesize();

	for (int fft = 0; fft < data_size; fft++) {
		out_buf_re[fft] = fft_obj.a[fft].real();
		out_buf_im[fft] = fft_obj.a[fft].imag();
	}
#endif

}

void multiply( int n, double *in1_re, double *in1_im, double *in2_re, double *in2_im, double *out_re, double *out_im ){
  double tmp_re;
  double tmp_im;
  int i;
  
  for( i = 0; i < n; i++ ){
    tmp_re =  in1_re[i] * in2_re[i] - in1_im[i] * in2_im[i];
    tmp_im =  in1_re[i] * in2_im[i] + in1_im[i] * in2_re[i];
    out_re[i] = tmp_re;
    out_im[i] = tmp_im;
  }
}

#define DISTANCE_OFFSET (-50) // margin for speaker-mic distance. 
int corr(double *x, int x_size, double *y, int y_size, int start_pos, int end_pos){

	double  *x_buf;
	double  *y_buf;
	double  *cxy; 
	double  s; 
	int     idx,idx2,search_len; 

	double max_cxy = 0;
	int    max_idx = 0;

    search_len = end_pos - start_pos;

	x_buf = (double*)malloc(search_len * 2 * sizeof(double));
	y_buf = (double*)malloc(search_len * 2 * sizeof(double));
	cxy   = (double*)malloc(search_len * sizeof(double));

	if(cxy==NULL){
		fprintf(stderr,"Can't_allocate_memory.\n");
		return 0;
	}

	memset(x_buf, 0, search_len * sizeof(double));
	memcpy(x_buf, &(x[start_pos]), search_len * sizeof(double));
	memset(y_buf, 0, search_len * sizeof(double));
	memcpy(y_buf, y, search_len * sizeof(double));

	for(idx = 0; idx < search_len-1; idx++){
		s = 0;
		for(idx2 = 0; idx2 < search_len-1; idx2++){
			s += x_buf[idx+idx2] * y_buf[idx2];
		}
		cxy[idx] = s;

		if(cxy[idx] > max_cxy){
			max_cxy = cxy[idx];
			max_idx = idx;
		}
	}

	printf("searched offset = %d\n", start_pos + max_idx);
	printf("suggest start offset = %d\n", start_pos + max_idx + DISTANCE_OFFSET);

	free(cxy);
	return start_pos + max_idx + DISTANCE_OFFSET;
}

int corr_on_fft(double *x, int x_size, double *y, int y_size, int start_pos, int end_pos){

	double  *x_buf;
	double  *x_fft_re, *x_fft_im;
	double  *y_buf;
	double  *y_fft_re, *y_fft_im;
	double  *fft_res_re, *fft_res_im;
	double  *res_buf_re, *res_buf_im;
	double  *cxy; 
	int     peak_corr_idx[10];
	int     search_len, len_temp; 

	double max_cxy = 0;
	int    max_idx = 0;

	int    fft_len_log, fft_len;

	search_len = end_pos - start_pos;

	if(search_len < y_size){
		len_temp = y_size;
	}else{
		len_temp = search_len;
	}

	fft_len_log = (int)floor(log(len_temp)/log(2));
	fft_len = (int)pow(2, fft_len_log); /* for fft */

	x_buf      = (double*)malloc(search_len * sizeof(double));
	x_fft_re   = (double*)malloc(fft_len * sizeof(double));
	x_fft_im   = (double*)malloc(fft_len * sizeof(double));
	y_buf      = (double*)malloc(y_size * sizeof(double));
	y_fft_re   = (double*)malloc(fft_len * sizeof(double));
	y_fft_im   = (double*)malloc(fft_len * sizeof(double));
	fft_res_re = (double*)malloc(fft_len * sizeof(double));
	fft_res_im = (double*)malloc(fft_len * sizeof(double));
	res_buf_re = (double*)malloc(fft_len * sizeof(double));
	res_buf_im = (double*)malloc(fft_len * sizeof(double));
	cxy        = (double*)malloc(fft_len * sizeof(double));

	if((x_buf==NULL) || (x_fft_re==NULL) || (x_fft_im==NULL) || (y_buf==NULL) || (y_fft_re==NULL) || (y_fft_im==NULL) 
	   || (fft_res_re==NULL) || (fft_res_im==NULL) || (res_buf_re==NULL) || (res_buf_im==NULL) || (cxy==NULL)){
		fprintf(stderr,"Can't_allocate_memory.\n");
		return 0;
	}

	memset(x_buf, 0, search_len * sizeof(double));
	memset(x_fft_re, 0, fft_len * sizeof(double));
	memset(x_fft_im, 0, fft_len * sizeof(double));
	memcpy(x_buf, &(x[start_pos]), search_len * sizeof(double));
	memset(y_buf, 0, y_size * sizeof(double));
	memset(y_fft_re, 0, fft_len * sizeof(double));
	memset(y_fft_im, 0, fft_len * sizeof(double));
	memcpy(y_buf, y, y_size * sizeof(double));

	memset(fft_res_re, 0, fft_len * sizeof(double));
	memset(fft_res_im, 0, fft_len * sizeof(double));
	memset(res_buf_re, 0, fft_len * sizeof(double));
	memset(res_buf_im, 0, fft_len * sizeof(double));

	memset(peak_corr_idx, 0, 10);

	fft(x_buf, x_fft_re, x_fft_im, fft_len);
	fft(y_buf, y_fft_re, y_fft_im, fft_len);

	free(x_buf);
	free(y_buf);
	

	for(int idx = 0; idx < fft_len-1; idx++){
		y_fft_im[idx] = -y_fft_im[idx];
	}

	multiply(fft_len, y_fft_re, y_fft_im, x_fft_re, x_fft_im, fft_res_re, fft_res_im);;

	ifft(fft_res_re, fft_res_im, res_buf_re, res_buf_im, fft_len);

	for(int idx = 0; idx < fft_len-1; idx++){
	
		cxy[idx]= sqrt((res_buf_re[idx] * res_buf_re[idx]) + (res_buf_im[idx] * res_buf_im[idx]));

		if(cxy[idx] > max_cxy){
			max_cxy = cxy[idx];
			max_idx = idx;
				
			for(int corr_idx = 9; corr_idx > 0; corr_idx--){
				peak_corr_idx[corr_idx] = peak_corr_idx[corr_idx-1];
			}
			peak_corr_idx[0] = max_idx;
		}
			
	}

	for(int corr_idx = 0; corr_idx < 10; corr_idx++){
		IFDEBUG(printf("10 peak %d:%d:%f\n", corr_idx, peak_corr_idx[corr_idx], cxy[peak_corr_idx[corr_idx]]));
		if( (peak_corr_idx[corr_idx]  < (max_idx - (16384 /4) * 3))
		    &&  (cxy[peak_corr_idx[corr_idx]] > (cxy[max_idx]/2))){
			IFDEBUG(printf("max_idx,peak_idx %d:%d\n", max_idx, peak_corr_idx[corr_idx]));
			max_idx = peak_corr_idx[corr_idx];
		}
		if(peak_corr_idx[corr_idx] == 0){
			return 0;
		}
	}

	printf("searched offset = %d\n", start_pos + max_idx);
	printf("suggest start offset = %d\n", start_pos + max_idx + DISTANCE_OFFSET);

	free(x_fft_re);
	free(x_fft_im);
	free(y_fft_re);
	free(y_fft_im);
	free(fft_res_re);
	free(fft_res_im);
	free(res_buf_re);
	free(res_buf_im);
	free(cxy);
	
	return start_pos + max_idx + DISTANCE_OFFSET;
}

#define TYPE_FLT_FILE 0
#define TYPE_WAV_FILE 1
#define MAX_CH 64
int main(int argc, char **argv) {
	int frame = 16384;
	char tsp_org_name[FILENAME_MAX];
	char tsp_org_type_flg = TYPE_FLT_FILE;
	char tsp_org_channel  = 0;
	int tsp_file_size = 0;
	double tsp_adjust_ratio = 32768;
	int tsp_adjust_ratio_set_flag= 0;
	int sync_add_freq = 16;
	int sync_add_offset = 16384;

	char search_offset_flag = 0;
	int  search_start_pos   = 0;
	int  search_end_pos     = 65536;

	char out_channel_flag[MAX_CH]; //flag of proccessing channel
	char out_channel_set_flag = 0;
	char enable_HPF_flg = 0;
	int enable_HPF_cutLow_freq = 0;
	int enable_HPF_cutHigh_freq = 0;

	double *tsp_wave_buf = NULL;
	double **tsp_wave;
	int tsp_length;

	char *i_wave_name;
	char *o_wave_name_orig;
	char o_wave_name[FILENAME_MAX];
	int h_size = 4096;
	int ch_idx = 0;
	int out_ch_idx = 0;

	int c;

	strcpy(tsp_org_name, "16384.little_endian.tsp");

	memset(out_channel_flag, 1, sizeof(char) * MAX_CH);

	IFDEBUG(FILE *debug_fd = fopen("debug.txt", "w+t"));

	if (argc < 3){
		usage();  // exit in funcition.
	}

	while (1) {
		//int this_option_optind = optind ? optind : 1;
		int option_index = 0;
		static struct option long_options[] = {
			{"tsporigflt", 1, 0, 0},
			{"tsporigwav", 1, 0, 0},
			{"tsporigchannel", 1, 0, 0},
			{"sigmax", 1, 0, 0},
			{"samplebyframe", 1, 0, 0},
			{"syncaddfreq", 1, 0, 0},
			{"syncaddoffset", 1, 0, 0},
			{"searchoffset", 1, 0, 0},
			{"outchannellist", 1, 0, 0},
			{"enableHPF", 1, 0, 0},
			{0, 0, 0, 0}
		};

		c = getopt_long (argc, argv, "v", long_options, &option_index);

		if (c == -1){
			break;
		}

		switch (c) {
			case 0:
				if(!strcmp(long_options[option_index].name, "tsporigflt")){
					strcpy(tsp_org_name, optarg);
					tsp_org_type_flg = TYPE_FLT_FILE;

				}else if(!strcmp(long_options[option_index].name, "tsporigwav")){
					strcpy(tsp_org_name, optarg);
					tsp_org_type_flg = TYPE_WAV_FILE;

				}else if(!strcmp(long_options[option_index].name, "tsporigchannel")){
					tsp_org_channel = atoi(optarg);

				}else if(!strcmp(long_options[option_index].name, "sigmax")){
					tsp_adjust_ratio = atoi(optarg);
					if(tsp_adjust_ratio == 0){
						printf("Adjust ratio must not 0.\n");
						exit(-1);
					}
					tsp_adjust_ratio_set_flag = 1;

				}else if(!strcmp(long_options[option_index].name, "samplebyframe")){
					frame = atoi(optarg);
					if(frame == 0){
						printf("Value of samplebyframe is incorrect. \n");
						exit(-1);
					}

				}else if(!strcmp(long_options[option_index].name, "syncaddfreq")){
					sync_add_freq = atoi(optarg);
					if(sync_add_freq == 0){
						printf("Value of syncaddfreq is incorrect. \n");
						exit(-1);
					}

				}else if(!strcmp(long_options[option_index].name, "syncaddoffset")){
					sync_add_offset = atoi(optarg);
				}else if(!strcmp(long_options[option_index].name, "searchoffset")){
					char *ch_str;

					ch_str = strtok(optarg, ",");
					if (ch_str != NULL) {
						search_start_pos = atoi(ch_str);
						ch_str = strtok(NULL, "\n");
						search_end_pos = atoi(ch_str);
					}
					printf("search pos = %d-%d. \n", search_start_pos, search_end_pos);
					search_offset_flag = 1;

				}else if(!strcmp(long_options[option_index].name, "outchannellist")){
					char *ch_str;
					char *endptr = NULL;
					long int i;
					memset(out_channel_flag, 0, sizeof(char) * MAX_CH);

					ch_str = strtok(optarg, ",");
					while (ch_str != NULL) {
						i = strtol(ch_str, &endptr, 10);
						if (ch_str != endptr && 0 <= i && i < MAX_CH) {
							out_channel_flag[i] = 1;
						}
						ch_str = strtok(NULL, ",");
					}
					out_channel_set_flag = 1;
				}else if(!strcmp(long_options[option_index].name, "enableHPF")){
					char *ch_str;
					char *endptr = NULL;
					long int i;

					enable_HPF_flg = true;

					ch_str = strtok(optarg, ",");
					i = strtol(ch_str, &endptr, 10);
					if (ch_str != endptr && 0 <= i ) {
						enable_HPF_cutLow_freq = i;
					}
					ch_str = strtok(NULL, ",");
					i = strtol(ch_str, &endptr, 10);
					if (ch_str != endptr && 0 <= i ) {
						enable_HPF_cutHigh_freq = i;
					}
				}
				
				break;

			case 'v':
				printf("calc impluse from TSP wave(compile time: %s %s)\n",__DATE__ ,__TIME__);
				exit(1);

			default:
				printf ("?? getopt returned character code 0%o ??\n", c);
				exit(-1);
		}
	}

	if((argc - optind) != 2){
		printf ("Error: incorrect arg count.\n");
		usage();  // exit in funcition.
	}

	if(search_offset_flag && (out_channel_set_flag == 0)){ //search_offset needs out_channel_list
		printf ("Error: --searchoffset needs --outchannellist for calculation.\n");
		usage();  // exit in funcition.
	}

	i_wave_name      = argv[optind++];
	o_wave_name_orig = argv[optind++];

	// TSP波形読み込み
	double *tsp_fft_re = (double *) malloc(sizeof(double) * frame);
	double *tsp_fft_im = (double *) malloc(sizeof(double) * frame);
	int frame_cnt;

	if(tsp_org_type_flg == TYPE_FLT_FILE){
		tsp_file_size = get_flt_size(tsp_org_name);
		tsp_length    = tsp_file_size /sizeof(float);

		printf("tsp_file_size = %d\n", tsp_file_size);

		tsp_wave_buf = (double *) malloc(tsp_file_size/ sizeof(float) * sizeof(double));

		if(tsp_adjust_ratio_set_flag){
			if(read_flt(tsp_org_name, tsp_wave_buf, tsp_file_size) == -1){
				printf("tsp origin file is not read.\n");
				exit(-1);
			}
		}else{
			if(read_flt(tsp_org_name, tsp_wave_buf, tsp_file_size, &tsp_adjust_ratio) == -1){
				printf("tsp origin file is not read.\n");
				exit(-1);
			}
		}

		printf("tsp_adjust_ratio = %f\n", tsp_adjust_ratio);

		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			tsp_wave_buf[frame_cnt] /= tsp_adjust_ratio;
		}

		//fft(tsp_wave_buf, tsp_fft, tsp_file_size / sizeof(float) * sizeof(double));
		fft(tsp_wave_buf, tsp_fft_re, tsp_fft_im, frame);

	}else if(tsp_org_type_flg == TYPE_WAV_FILE){
		int tsp_ch;
		int tsp_samplerate;

		if(readwavfile(tsp_org_name, &tsp_wave, &tsp_length, &tsp_ch, &tsp_samplerate) == -1){
			printf("wavefile(%s) is not found.\n", i_wave_name);
			exit(-1);
		}

		fft(tsp_wave[tsp_org_channel], tsp_fft_re, tsp_fft_im, frame);

	}else{
		printf ("Error: incorrect tsp_org_type.\n");
		usage();  // exit in funcition.
	}

	IFDEBUG(fprintf(debug_fd, "/////TSP FFT/////\n");
	for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
		fprintf(debug_fd,"%f+%fi\n", tsp_fft_re[frame_cnt], tsp_fft_im[frame_cnt]);
	}
	fprintf(debug_fd, "\n\n\n\n");
		)

	IFDEBUG(fflush(debug_fd));


	// マイク録音読み込み
	double **iwave;
	int ilength;
	int input_ch;
	int isamplerate;

	if(readwavfile(i_wave_name, &iwave, &ilength, &input_ch, &isamplerate) == -1){
		printf("wavefile(%s) is not found.\n", i_wave_name);
		exit(-1);
	}

	if(((sync_add_freq * frame) + sync_add_offset) > ilength){
		printf("Wave length(%d) is short when sync_add_freq = %d.\n", ilength, sync_add_freq);
		exit(-1);
	}

	IFDEBUG(
		fprintf(debug_fd, "/////INPUT WAVE/////\n");
		for (frame_cnt = 0; frame_cnt < /*frame*/ 1024; frame_cnt++) {
			fprintf(debug_fd, "%f\n", iwave[0][frame_cnt]);
		}
		fprintf(debug_fd, "\n\n\n\n");
	)

	IFDEBUG(fclose(debug_fd));
	IFDEBUG(debug_fd = fopen("debug.txt", "a+t"));

	// インパルス応答計算
	printf("calc_impulse...\n");
	double *xsum = (double *) malloc(sizeof(double) * frame);
	double *xmean = (double *) malloc(sizeof(double) * frame);
	double *xfft_re = (double *) malloc(sizeof(double) * frame);
	double *xfft_im = (double *) malloc(sizeof(double) * frame);
	double *h_wave_buf_re = (double *) malloc(sizeof(double) * frame);
	double *h_wave_buf_im = (double *) malloc(sizeof(double) * frame);
	double *hfft_re = (double *) malloc(sizeof(double) * frame);
	double *hfft_im = (double *) malloc(sizeof(double) * frame);
	double *h_wave_short_buf_re = (double *) malloc(sizeof(double) * h_size);
	double *h_wave_short_buf_im = (double *) malloc(sizeof(double) * h_size);
	double *hhpf_re = (double *) malloc(sizeof(double) * h_size);
	double *hhpf_im = (double *) malloc(sizeof(double) * h_size);
	double *hpf_coef = (double *) malloc(sizeof(double) * h_size);

	if(search_offset_flag){
		printf("search_offset...\n");
		int calculated_offset, calculated_first_offset = search_end_pos;
		//オフセット推定処理
		for (ch_idx = 0, out_ch_idx = 0; ch_idx < input_ch; ch_idx++) {
			if (out_channel_flag[ch_idx] == 0){
				continue;
			}
		
			if(search_end_pos > ilength){
				search_end_pos = ilength;
			}

			// 探索範囲がTSPのサイズより狭かった場合、TSPのサイズまで広げる
			if( (search_end_pos - search_start_pos) < tsp_length){
				if( (search_start_pos - tsp_length) >= 0){
					search_start_pos -= tsp_length;
					printf("INFO: Search length is under tsp size. So changing search start pos = %d.\n", search_start_pos );
				}else if( ( search_end_pos + tsp_length) <= ilength ){
					search_end_pos += tsp_length;
					printf("INFO: Search length is under tsp size. So changing search end pos = %d.\n", search_end_pos );
				}else{
					printf("ERROR: Search length is under tsp size. But there is not enough area on impulse.\n");
					exit(-1);
				}
			}

			if(tsp_org_type_flg == TYPE_FLT_FILE){
				//sync_add_offset = corr(iwave[0], ilength, tsp_wave_buf, tsp_file_size, search_start_pos, search_end_pos);
				calculated_offset = corr_on_fft(iwave[ch_idx], ilength, tsp_wave_buf, tsp_length, search_start_pos, search_end_pos);
			}else{
				//sync_add_offset = corr(iwave[0], ilength, tsp_wave[tsp_org_channel], tsp_length, search_start_pos, search_end_pos);
				calculated_offset = corr_on_fft(iwave[ch_idx], ilength, tsp_wave[tsp_org_channel], tsp_length, search_start_pos, search_end_pos);
			}
			if(calculated_offset < calculated_first_offset){
				calculated_first_offset = calculated_offset;
			}
		}
		printf("calculated_first_offset=%d.\n", calculated_first_offset);
		sync_add_offset = calculated_first_offset;
	}

	for (ch_idx = 0, out_ch_idx = 0; ch_idx < input_ch; ch_idx++) {
		int frame_idx;
		
		if (out_channel_flag[ch_idx] == 0){
			continue;
		}

		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			xsum[frame_cnt] = 0;

			for (frame_idx = 0; frame_idx < sync_add_freq; frame_idx++) {
				xsum[frame_cnt] += iwave[ch_idx][frame_idx * frame + frame_cnt + sync_add_offset];
			}
		}
		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			xmean[frame_cnt] = xsum[frame_cnt] / sync_add_freq;
		}

		fft(xmean, xfft_re, xfft_im, frame);

		IFDEBUG(fprintf(debug_fd, "/////X FFT/////\n");
		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			fprintf(debug_fd, "%f+%fi\n", xfft_re[frame_cnt], xfft_im[frame_cnt]);
		}
		fprintf(debug_fd, "\n\n\n\n");
		fprintf(debug_fd, "fft end...\n");
			)
		

		double denomi = 0.0;
		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			denomi = tsp_fft_re[frame_cnt] * tsp_fft_re[frame_cnt] + tsp_fft_im[frame_cnt] * tsp_fft_im[frame_cnt];
			hfft_re[frame_cnt] = (xfft_re[frame_cnt] * tsp_fft_re[frame_cnt] + xfft_im[frame_cnt] * tsp_fft_im[frame_cnt])
					/ denomi;
			hfft_im[frame_cnt] = (tsp_fft_re[frame_cnt] * xfft_im[frame_cnt] - xfft_re[frame_cnt] * tsp_fft_im[frame_cnt])
					/ denomi;
		}

		IFDEBUG(
		fprintf(debug_fd, "/////H FFT/////\n");
		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			fprintf(debug_fd, "%f+%fi\n", hfft_re[frame_cnt], hfft_im[frame_cnt]);
		}
		fprintf(debug_fd, "\n\n\n\n");
		)
		ifft(hfft_re, hfft_im, h_wave_buf_re, h_wave_buf_im, frame);


		for (frame_cnt = 0; frame_cnt < frame; frame_cnt++) {
			h_wave_buf_re[frame_cnt] /= frame;
			h_wave_buf_im[frame_cnt] /= frame;
		}

		// 初期フレームだけ切り出し
		for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
			h_wave_short_buf_re[frame_cnt] = h_wave_buf_re[frame_cnt];
			h_wave_short_buf_im[frame_cnt] = h_wave_buf_im[frame_cnt];
		}

		IFDEBUG(
		fprintf(debug_fd, "/////h short wave/////\n");
		for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
			fprintf(debug_fd, "%f+%fi\n", h_wave_short_buf_re[frame_cnt], h_wave_short_buf_im[frame_cnt]);
		}
		fprintf(debug_fd, "\n\n\n\n");
		)

		if(enable_HPF_flg){

			fft(h_wave_short_buf_re, hhpf_re, hhpf_im, h_size);

			IFDEBUG(
			fprintf(debug_fd, "/////before HPF/////\n");
			for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
				fprintf(debug_fd, "%f+%fi\n", hhpf_re[frame_cnt], hhpf_im[frame_cnt]);
			}
			fprintf(debug_fd, "\n\n\n\n");
			)

			// 低周波カット
			double cut_f1 = enable_HPF_cutLow_freq;
			double cut_f2 = enable_HPF_cutHigh_freq;
			double fs = 16000;
			int cut1 = round(cut_f1 * h_size / fs);
			int cut2 = round(cut_f2 * h_size / fs);
			double cut_d = cut2 - cut1;

			for (frame_cnt = 0; frame_cnt < h_size / 2 + 1; frame_cnt++) {
				if (frame_cnt < cut1) {
					hpf_coef[frame_cnt] = 0;
					hpf_coef[h_size - frame_cnt - 1] = 0;
				} else if (frame_cnt < cut2) {
					hpf_coef[frame_cnt] = (frame_cnt + 1 - cut1) / cut_d;
					hpf_coef[h_size - frame_cnt] = (frame_cnt + 1 - cut1) / cut_d;
				} else {
					hpf_coef[frame_cnt] = 1;
					hpf_coef[h_size - frame_cnt] = 1;
				}
			}

			IFDEBUG(
			fprintf(debug_fd, "/////HPF COEF/////\n");
			for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
				fprintf(debug_fd, "%f\n", hpf_coef[frame_cnt]);
			}
			fprintf(debug_fd, "cutHz=%f:%f", cut_f1, cut_f2);
			fprintf(debug_fd, "\n\n\n\n");
			)

			for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
				hhpf_re[frame_cnt] = hhpf_re[frame_cnt] * hpf_coef[frame_cnt];
				hhpf_im[frame_cnt] = hhpf_im[frame_cnt] * hpf_coef[frame_cnt];
			}

			IFDEBUG(
			fprintf(debug_fd, "/////after HPF/////\n");
			for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
				fprintf(debug_fd, "%f+%fi\n", hhpf_re[frame_cnt], hhpf_im[frame_cnt]);
			}
			fprintf(debug_fd, "\n\n\n\n");
			)

			ifft(hhpf_re, hhpf_im, h_wave_short_buf_re, h_wave_short_buf_im, h_size);

			for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
				h_wave_short_buf_re[frame_cnt] /= h_size;
				h_wave_short_buf_im[frame_cnt] /= h_size;
			}

			IFDEBUG(
			fprintf(debug_fd, "/////short wave after HPF/////\n");
			for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
				fprintf(debug_fd, "%f+%fi\n", h_wave_short_buf_re[frame_cnt], h_wave_short_buf_im[frame_cnt]);
			}
			fprintf(debug_fd, "\n\n\n\n");
			)
		}

		sprintf(o_wave_name, "%s_ch%d.flt", o_wave_name_orig, out_ch_idx + 1);
		if(write_flt(o_wave_name, h_wave_short_buf_re, h_size) == -1){
		  fprintf(stderr, "File open failed: %s\n", o_wave_name);
		  exit(1);
		}
		
		out_ch_idx++;
	}

	IFDEBUG(printf("all function is end...\n"));

	free(hhpf_re);
	free(hhpf_im);
	free(hpf_coef);
	free(h_wave_short_buf_re);
	free(h_wave_short_buf_im);
	free(hfft_re);
	free(hfft_im);
	free(h_wave_buf_re);
	free(h_wave_buf_im);
	free(xfft_re);
	free(xfft_im);
	free(xmean);
	free(xsum);
	free(tsp_fft_re);
	free(tsp_fft_im);
	free(tsp_wave_buf);

	for(c=0; c<input_ch; c++)
	  free(iwave[c]);
	free(iwave);

	IFDEBUG(fclose(debug_fd));

	return 0;
}

